@@ -27,6 +27,14 @@
 #include <ctype.h>
 #include <math.h>
 #include <assert.h>
+#include <signal.h>
+#include <sys/time.h>
+
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#else
+#define VERSION "unknown"
+#endif
 
 #include "simpleaudio.h"
 #include "fsk.h"
@@ -80,11 +88,38 @@ framebits_decode_baudot( char *dataout_p, unsigned int dataout_size,
 }
 
 
+
+int		tx_transmitting = 0;
+int		tx_leader_bits_len = 2;
+int		tx_trailer_bits_len = 2;
+
+simpleaudio	*tx_sa_out;
+float		tx_bfsk_mark_f;
+unsigned int	tx_bit_nsamples;
+
+void
+tx_stop_transmit_sighandler( int sig )
+{
+    // fprintf(stderr, "alarm\n");
+
+    int j;
+    for ( j=0; j<tx_trailer_bits_len; j++ )
+	simpleaudio_tone(tx_sa_out, tx_bfsk_mark_f, tx_bit_nsamples);
+
+    // 0.5 sec of zero samples to flush - FIXME lame
+    size_t sample_rate = simpleaudio_get_rate(tx_sa_out);
+    simpleaudio_tone(tx_sa_out, 0, sample_rate/2);
+
+    tx_transmitting = 0;
+}
+
+
 /*
  * rudimentary BFSK transmitter
  */
 static void fsk_transmit_stdin(
 	simpleaudio *sa_out,
+	int tx_interactive,
 	float data_rate,
 	float bfsk_mark_f,
 	float bfsk_space_f,
@@ -97,15 +132,37 @@ static void fsk_transmit_stdin(
     size_t bit_nsamples = sample_rate / data_rate + 0.5;
     int c;
 
-    simpleaudio_tone(sa_out, bfsk_mark_f, bit_nsamples);    // 2.0 bit leader
-    simpleaudio_tone(sa_out, bfsk_mark_f, bit_nsamples);
+    tx_sa_out = sa_out;
+    tx_bfsk_mark_f = bfsk_mark_f;
+    tx_bit_nsamples = bit_nsamples;
+
+    // one-shot
+    struct itimerval itv = {
+	{0, 0},						// it_interval
+	{0, 1000000/(float)(data_rate+data_rate*0.03) }	// it_value
+    };
+
+    if ( tx_interactive )
+	signal(SIGALRM, tx_stop_transmit_sighandler);
+
+    tx_transmitting = 0;
     while ( (c = getchar()) != EOF )
     {
+	if ( tx_interactive )
+	    setitimer(ITIMER_REAL, NULL, NULL);
 
+	// fprintf(stderr, "<c=%d>", c);
 	unsigned int nwords;
 	unsigned int bits[2];
 	nwords = framebits_encoder(bits, c);
 
+	if ( !tx_transmitting )
+	{
+	    tx_transmitting = 1;
+	    int j;
+	    for ( j=0; j<tx_leader_bits_len; j++ )
+		simpleaudio_tone(sa_out, bfsk_mark_f, bit_nsamples);
+	}
 	unsigned int j;
 	for ( j=0; j<nwords; j++ ) {
 	    simpleaudio_tone(sa_out, bfsk_space_f, bit_nsamples);	// start
@@ -118,13 +175,18 @@ static void fsk_transmit_stdin(
 	    simpleaudio_tone(sa_out, bfsk_mark_f,
 				bit_nsamples * bfsk_txstopbits);	// stop
 	}
+
+	if ( tx_interactive )
+	    setitimer(ITIMER_REAL, &itv, NULL);
     }
-    simpleaudio_tone(sa_out, bfsk_mark_f, bit_nsamples);    // 2.0 bit tail
-    simpleaudio_tone(sa_out, bfsk_mark_f, bit_nsamples);
-
-    // 0.5 sec of zero samples to flush - FIXME lame
-    simpleaudio_tone(sa_out, 0, sample_rate/2);
+    if ( tx_interactive ) {
+	setitimer(ITIMER_REAL, NULL, NULL);
+	signal(SIGALRM, SIG_DFL);
+    }
+    if ( !tx_transmitting )
+	return;
 
+    tx_stop_transmit_sighandler(0);
 }
 
 
@@ -132,23 +194,47 @@ static void
 report_no_carrier( fsk_plan *fskp,
 	unsigned int sample_rate,
 	float bfsk_data_rate,
+	float nsamples_per_bit,
 	unsigned int nframes_decoded,
 	size_t carrier_nsamples,
 	float confidence_total )
 {
     unsigned long long nbits_total = nframes_decoded * (fskp->n_data_bits+2);
+#if 0
+    fprintf(stderr, "nframes_decoded=%u\n", nframes_decoded);
+    fprintf(stderr, "nbits_total=%llu\n", nbits_total);
+    fprintf(stderr, "carrier_nsamples=%lu\n", carrier_nsamples);
+    fprintf(stderr, "nsamples_per_bit=%f\n", nsamples_per_bit);
+#endif
     float throughput_rate = nbits_total * sample_rate / (float)carrier_nsamples;
-    float throughput_skew = (throughput_rate - bfsk_data_rate)
-			/ bfsk_data_rate;
-    fprintf(stderr, "### NOCARRIER ndata=%u confidence=%.2f throughput=%.2f (%.1f%% %s) ###\n",
+    fprintf(stderr, "### NOCARRIER ndata=%u confidence=%.2f throughput=%.2f",
 	    nframes_decoded,
 	    confidence_total / nframes_decoded,
-	    throughput_rate,
-	    fabs(throughput_skew) * 100.0,
-	    signbit(throughput_skew) ? "slow" : "fast"
-	    );
+	    throughput_rate);
+    if ( (size_t)(nbits_total * nsamples_per_bit + 0.5) == carrier_nsamples ) {
+	fprintf(stderr, " (rate perfect) ###\n");
+    } else {
+	float throughput_skew = (throughput_rate - bfsk_data_rate)
+			    / bfsk_data_rate;
+	fprintf(stderr, " (%.1f%% %s) ###\n",
+		fabs(throughput_skew) * 100.0,
+		signbit(throughput_skew) ? "slow" : "fast"
+		);
+    }
 }
 
+void
+version()
+{
+    printf(
+    "minimodem %s\n"
+    "Copyright (C) 2011 Kamal Mostafa <kamal@whence.com>\n"
+    "License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>.\n"
+    "This is free software: you are free to change and redistribute it.\n"
+    "There is NO WARRANTY, to the extent permitted by law.\n\n"
+    "Written by Kamal Mostafa <kamal@whence.com>.\n",
+	VERSION);
+}
 
 void
 usage()
@@ -158,6 +244,8 @@ usage()
     "		    -t, --tx, --transmit, --write\n"
     "		    -r, --rx, --receive,  --read     (default)\n"
     "		[options]\n"
+    "		    -a, --auto-carrier\n"
+    "		    -c, --confidence {threshold}\n"
     "		    -8, --ascii		ASCII  8-N-1\n"
     "		    -5, --baudot	Baudot 5-N-1\n"
     "		    -f, --file {filename.flac}\n"
@@ -165,9 +253,11 @@ usage()
     "		    -M, --mark {mark_freq}\n"
     "		    -S, --space {space_freq}\n"
     "		    -T, --txstopbits {m.n}\n"
+    "		    -q, --quiet\n"
+    "		    -V, --version\n"
     "		{baudmode}\n"
     "		    1200 : Bell202  1200 bps --ascii\n"
-    "		     300 : Bell103   300 bps --ascii (auto-rx-carrier)\n"
+    "		     300 : Bell103   300 bps --ascii\n"
     "		  N>=100 : Bellxxx     N bps --ascii\n"
     "		   N<100 : RTTY        N bps --baudot\n"
     "		    rtty : RTTY    45.45 bps --baudot\n"
@@ -180,15 +270,19 @@ main( int argc, char*argv[] )
 {
     char *modem_mode = NULL;
     int TX_mode = -1;
+    int quiet_mode = 0;
     float band_width = 0;
     unsigned int bfsk_mark_f = 0;
     unsigned int bfsk_space_f = 0;
     float bfsk_txstopbits = 0;
     unsigned int bfsk_n_data_bits = 0;
-    unsigned int autodetect_shift;
+    int autodetect_shift;
     char *filename = NULL;
     char *program_name;
 
+    float	carrier_autodetect_threshold = 0.0;
+    float	bfsk_confidence_threshold = 0.6;
+
     program_name = strrchr(argv[0], '/');
     if ( program_name )
 	program_name++;
@@ -200,12 +294,15 @@ main( int argc, char*argv[] )
     
     while ( 1 ) {
 	static struct option long_options[] = {
+	    { "version",	0, 0, 'V' },
 	    { "tx",		0, 0, 't' },
 	    { "transmit",	0, 0, 't' },
 	    { "write",		0, 0, 't' },
 	    { "rx",		0, 0, 'r' },
 	    { "receive",	0, 0, 'r' },
 	    { "read",		0, 0, 'r' },
+	    { "confidence",	1, 0, 'c' },
+	    { "auto-carrier",	0, 0, 'a' },
 	    { "ascii",		0, 0, '8' },
 	    { "baudot",		0, 0, '5' },
 	    { "file",		1, 0, 'f' },
@@ -213,13 +310,17 @@ main( int argc, char*argv[] )
 	    { "mark",		1, 0, 'M' },
 	    { "space",		1, 0, 'S' },
 	    { "txstopbits",	1, 0, 'T' },
+	    { "quiet",		0, 0, 'q' },
 	    { 0 }
 	};
-	c = getopt_long(argc, argv, "tr85f:b:M:S:T:",
+	c = getopt_long(argc, argv, "Vtrc:a85f:b:M:S:T:q",
 		long_options, &option_index);
 	if ( c == -1 )
 	    break;
 	switch( c ) {
+	    case 'V':
+			version();
+			exit(0);
 	    case 't':
 			if ( TX_mode == 0 )
 			    usage();
@@ -230,6 +331,12 @@ main( int argc, char*argv[] )
 			    usage();
 			TX_mode = 0;
 			break;
+	    case 'c':
+			bfsk_confidence_threshold = atof(optarg);
+			break;
+	    case 'a':
+			carrier_autodetect_threshold = 0.001;
+			break;
 	    case 'f':
 			filename = optarg;
 			break;
@@ -255,6 +362,9 @@ main( int argc, char*argv[] )
 			bfsk_txstopbits = atof(optarg);
 			assert( bfsk_txstopbits > 0 );
 			break;
+	    case 'q':
+			quiet_mode = 1;
+			break;
 	    default:
 			usage();
 	}
@@ -289,6 +399,8 @@ main( int argc, char*argv[] )
 	bfsk_data_rate = 45.45;
 	if ( bfsk_n_data_bits == 0 )
 	    bfsk_n_data_bits = 5;
+	if ( bfsk_txstopbits == 0 )
+	    bfsk_txstopbits = 1.5;
     } else {
 	bfsk_data_rate = atof(modem_mode);
 	if ( bfsk_n_data_bits == 0 )
@@ -307,16 +419,15 @@ main( int argc, char*argv[] )
 	assert( 0 && bfsk_n_data_bits );
     }
 
-
     if ( bfsk_data_rate >= 400 ) {
 	/*
 	 * Bell 202:     baud=1200 mark=1200 space=2200
 	 */
-	autodetect_shift = 0;	// not used
+	autodetect_shift = - ( bfsk_data_rate * 5 / 6 );
 	if ( bfsk_mark_f == 0 )
-	    bfsk_mark_f  = 1200;
+	    bfsk_mark_f  = bfsk_data_rate / 2 + 600;
 	if ( bfsk_space_f == 0 )
-	    bfsk_space_f = bfsk_mark_f + 1000;
+	    bfsk_space_f = bfsk_mark_f - autodetect_shift;
 	if ( band_width == 0 )
 	    band_width = 200;
     } else if ( bfsk_data_rate >= 100 ) {
@@ -340,8 +451,6 @@ main( int argc, char*argv[] )
 	    bfsk_mark_f  = 1585;
 	if ( bfsk_space_f == 0 )
 	    bfsk_space_f = bfsk_mark_f - autodetect_shift;
-	if ( bfsk_txstopbits == 0 )
-	    bfsk_txstopbits = 1.5;	// conventional for RTTY (?)
 	if ( band_width == 0 ) {
 	    band_width = 10;	// FIXME chosen arbitrarily
 	}
@@ -361,8 +470,10 @@ main( int argc, char*argv[] )
     if ( TX_mode ) {
 
 	simpleaudio *sa_out = NULL;
+	int tx_interactive = 1;
 
 	if ( filename ) {
+	    tx_interactive = 0;
 	    sa_out = simpleaudio_open_stream_sndfile(SA_STREAM_PLAYBACK,
 					filename);
 	    if ( ! sa_out )
@@ -371,9 +482,10 @@ main( int argc, char*argv[] )
 	if ( ! sa_out )
 	    sa_out = simpleaudio_open_stream_pulseaudio(SA_STREAM_PLAYBACK,
 					program_name, "output audio");
-	assert( sa_out );
+	if ( ! sa_out )
+	    return 1;
 
-	fsk_transmit_stdin(sa_out,
+	fsk_transmit_stdin(sa_out, tx_interactive,
 				bfsk_data_rate,
 				bfsk_mark_f, bfsk_space_f,
 				bfsk_n_data_bits,
@@ -496,12 +608,9 @@ main( int argc, char*argv[] )
 	if ( samples_nvalid == 0 )
 	    break;
 
-#define CARRIER_AUTODETECT_THRESHOLD	0.03
-#ifdef CARRIER_AUTODETECT_THRESHOLD
 	/* Auto-detect carrier frequency */
 	static int carrier_band = -1;
-	// FIXME?: hardcoded 300 baud trigger for carrier autodetect
-	if ( bfsk_data_rate <= 300 && carrier_band < 0 ) {
+	if ( carrier_autodetect_threshold > 0.0 && carrier_band < 0 ) {
 	    unsigned int i;
 	    float nsamples_per_scan = nsamples_per_bit;
 	    if ( nsamples_per_scan > fskp->fftsize )
@@ -510,7 +619,7 @@ main( int argc, char*argv[] )
 						 i+=nsamples_per_scan ) {
 		carrier_band = fsk_detect_carrier(fskp,
 				    samplebuf+i, nsamples_per_scan,
-				    CARRIER_AUTODETECT_THRESHOLD);
+				    carrier_autodetect_threshold);
 		if ( carrier_band >= 0 )
 		    break;
 	    }
@@ -538,7 +647,6 @@ main( int argc, char*argv[] )
 
 	    fsk_set_tones_by_bandshift(fskp, /*b_mark*/carrier_band, b_shift);
 	}
-#endif
 
 	/*
 	 * The main processing algorithm: scan samplesbuf for FSK frames,
@@ -553,7 +661,9 @@ main( int argc, char*argv[] )
 	    break;
 
 	unsigned int try_max_nsamples = nsamples_per_bit;
-#define FSK_ANALYZE_NSTEPS		8	/* accuracy vs. performance */
+#define FSK_ANALYZE_NSTEPS		10	/* accuracy vs. performance */
+		// Note: FSK_ANALYZE_NSTEPS has subtle effects on the
+		// "rate perfect" calculation.  oh well.
 	unsigned int try_step_nsamples = nsamples_per_bit / FSK_ANALYZE_NSTEPS;
 	if ( try_step_nsamples == 0 )
 	    try_step_nsamples = 1;
@@ -571,20 +681,29 @@ main( int argc, char*argv[] )
 			&frame_start_sample
 			);
 
-#define FSK_MIN_CONFIDENCE		0.05
+	// FIXME: hardcoded chop off framing bits
+	if ( fskp->n_data_bits == 5 )
+	    bits = ( bits >> 2 ) & 0x1F;
+	else
+	    bits = ( bits >> 2 ) & 0xFF;
+
 #define FSK_MAX_NOCONFIDENCE_BITS	20
 
-	if ( confidence <= FSK_MIN_CONFIDENCE ) {
+#define FSK_SCAN_LAG			0.2
+
+	if ( confidence <= bfsk_confidence_threshold ) {
 	    // FIXME: explain
 	    if ( ++noconfidence > FSK_MAX_NOCONFIDENCE_BITS )
 	    {
-#ifdef CARRIER_AUTODETECT_THRESHOLD
 		carrier_band = -1;
-#endif
 		if ( carrier ) {
 		    carrier_nsamples -= noconfidence_nsamples;
-		    report_no_carrier(fskp, sample_rate, bfsk_data_rate,
-			nframes_decoded, carrier_nsamples, confidence_total);
+		    if ( nframes_decoded > 0 )
+			carrier_nsamples += nsamples_per_bit * FSK_SCAN_LAG;
+		    if ( !quiet_mode )
+			report_no_carrier(fskp, sample_rate, bfsk_data_rate,
+			    nsamples_per_bit, nframes_decoded,
+			    carrier_nsamples, confidence_total);
 		    carrier = 0;
 		    carrier_nsamples = 0;
 		    confidence_total = 0;
@@ -601,15 +720,19 @@ main( int argc, char*argv[] )
 	}
 
 	if ( !carrier ) {
-	    if ( bfsk_data_rate >= 100 )
-		fprintf(stderr, "### CARRIER %u @ %.1f Hz ###\n",
-			(unsigned int)(bfsk_data_rate + 0.5),
-			fskp->b_mark * fskp->band_width);
-	    else
-		fprintf(stderr, "### CARRIER %.2f @ %.1f Hz ###\n",
-			bfsk_data_rate,
-			fskp->b_mark * fskp->band_width);
+	    if ( !quiet_mode ) {
+		if ( bfsk_data_rate >= 100 )
+		    fprintf(stderr, "### CARRIER %u @ %.1f Hz ###\n",
+			    (unsigned int)(bfsk_data_rate + 0.5),
+			    fskp->b_mark * fskp->band_width);
+		else
+		    fprintf(stderr, "### CARRIER %.2f @ %.1f Hz ###\n",
+			    bfsk_data_rate,
+			    fskp->b_mark * fskp->band_width);
+	    }
 	    carrier = 1;
+	    /* back up carrier_nsamples to account for the imminent advance */
+	    noconfidence_nsamples = frame_start_sample;
 	    bfsk_framebits_decode(0, 0, 0);	/* reset the frame processor */
 	}
 
@@ -617,7 +740,6 @@ main( int argc, char*argv[] )
 	confidence_total += confidence;
 	nframes_decoded++;
 	noconfidence = 0;
-	noconfidence_nsamples = 0;
 
 	/* Advance the sample stream forward past the decoded frame
 	 * but not past the stop bit, since we want it to appear as
@@ -628,7 +750,6 @@ main( int argc, char*argv[] )
 	 * but actually advance just a bit less than that to allow
 	 * for clock skew, hence FSK_SCAN_LAG.
 	 */
-#define FSK_SCAN_LAG			0.2
 	advance = frame_start_sample +
 	    nsamples_per_bit * (float)(fskp->n_data_bits + 2 - FSK_SCAN_LAG);
 
@@ -667,8 +788,12 @@ main( int argc, char*argv[] )
 
     if ( carrier ) {
 	carrier_nsamples -= noconfidence_nsamples;
-	report_no_carrier(fskp, sample_rate, bfsk_data_rate,
-	    nframes_decoded, carrier_nsamples, confidence_total);
+	if ( nframes_decoded > 0 )
+	    carrier_nsamples += nsamples_per_bit * FSK_SCAN_LAG;
+	if ( !quiet_mode )
+	    report_no_carrier(fskp, sample_rate, bfsk_data_rate,
+		nsamples_per_bit, nframes_decoded,
+		carrier_nsamples, confidence_total);
     }
 
     simpleaudio_close(sa);
